export function cors(options?: {
  allowMethods?: string[] | string,
  exposeHeaders?: string[] | string,
  allowHeaders?: string[] | string,
  origin?: Function | string,
  maxAge?: number | string,
  credentials?: boolean,
  keepHeadersOnError?: boolean
}) {
  const defaults = {
    allowMethods: 'GET,HEAD,PUT,POST,DELETE,PATCH'
  }

  options = options || {}
  options.allowMethods = defaults.allowMethods

  if (Array.isArray(options.exposeHeaders)) {
    options.exposeHeaders = options.exposeHeaders.join(',')
  }

  if (Array.isArray(options.allowMethods)) {
    options.allowMethods = options.allowMethods.join(',')
  }

  if (Array.isArray(options.allowHeaders)) {
    options.allowHeaders = options.allowHeaders.join(',')
  }

  if (options.maxAge) {
    options.maxAge = String(options.maxAge)
  }

  options.credentials = !!options.credentials

  options.keepHeadersOnError = options.keepHeadersOnError === undefined || !!options.keepHeadersOnError

  return async (ctx, next: Function) => {
    let requestOrigin = ctx.headers['origin']
    if (!requestOrigin) return await next()

    let origin

    if (typeof options.origin === 'function') {
      origin = await options.origin(ctx)
      if (!origin) return await next()
    } else {
      origin = options.origin || requestOrigin
    }

    let headers: {
      [key: string]: string
    } = {}

    if (ctx.req.method !== 'OPTIONS') {
      headers['Access-Control-Allow-Origin'] = origin
      if (options.credentials === true) {
        // If the origin is set to `*`, `credentials` can't be true, so log a warning
        // and clear the access-control-allow-credentials header.
        if (origin === '*') {
          // throw 'Invalid CORS settings for route `' + this.url + '`: if `origin` is \'*\', `credentials` cannot be `true` (setting `credentials` to `false` for you).')
          delete ctx.req.headers['Access-Control-Allow-Credentials']
        } else {
          // Otherwise set the access-control-allow-credentials header to `true`.
          headers['Access-Control-Allow-Credentials'] = 'true'
        }
      }
      if (options.exposeHeaders) headers['Access-Control-Expose-Headers'] = options.exposeHeaders as string
      if (!options.keepHeadersOnError) return await next()

      ctx.set(headers)

      await next()
    } else {
      if (!ctx.headers['access-control-request-method']) return await next()

      headers['Access-Control-Allow-Origin'] = origin
      if (options.credentials === true) headers['Access-Control-Allow-Credentials'] = 'true'
      if (options.maxAge) headers['Access-Control-Max-Age'] = options.maxAge as string
      if (options.allowMethods) headers['Access-Control-Allow-Methods'] = options.allowMethods as string

      let allowHeaders = options.allowHeaders
      if (!allowHeaders) allowHeaders = ctx.headers['access-control-request-headers']
      if (allowHeaders) headers['Access-Control-Allow-Headers'] = allowHeaders as string

      ctx.set(headers)

      ctx.routed = true
      ctx.status = 204
      requestOrigin = headers = origin = allowHeaders = null
    }
  }
}
